import java.io.*;

/**
 * INHX32 HEX to SYSEX converter to be used with MR-9090 bootloader.
 * <br/><br/>
 * Requires JRE v1.4+
 * <br/><br/>
 * Chunk sysex format:<br/>
 * <br/>
 * F7 71 55 17 7F AH AL &lt;64 bytes of midi encoded (BYTE%gt;&gt;7 BYTE&0x7F) data&gt; F7<br/>
 * <br/>
 *  AH = program memory address &gt;&gt; 7 & 0x7f<br/>
 *  AL = program memory address & 0x7f<br/>
 * <br/>
 *  so program memory address = AH&lt;&lt;7 | AL<br/>
 * <br/>
 * Note: supports only 14-bit addresses 0x0000 - 0x3FFF (18f442) and all chunks must be in 64-byte boundaries.<br/>
 * <br/>
 * Set delay (tested with 150ms) between sysex messages when dumping so flash writes get written<br/>
 *  correctly!<br/>
 * <br/>
 * Use 'java -jar syxconverter.jar prog.hex prog.syx' to generate the<br/>
 *  sysex file.<br/>
 *
 * <pre>
 * LICENSE
 * Creative Commons Attribution-NonCommercial-ShareALike 2.5
 *
 * You are free:
 *
 * * to copy, distribute, display, and perform the work
 * * to make derivative works
 *
 * Under the following conditions:
 *
 * 1. Attribution. You must attribute the work in the manner specified by the
 *    author or licensor.
 * 2. Noncommercial. You may not use this work for commercial purposes.
 * 3. Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under a license identical to this one.
 *
 * * For any reuse or distribution, you must make clear to others the license
 *   terms of this work.
 * * Any of these conditions can be waived if you get permission from the
 *   copyright holder.
 *
 * Your fair use and other rights are in no way affected by the above.
 *
 * This is a human-readable summary of the Legal Code.
 * See full license at http://creativecommons.org/licenses/by-nc-sa/2.5/legalcode
 *
 * Copyright: Marko Kanala (raato@mulletronic.com).
 * </pre>
 *
 * @author Marko Kanala, raato@mulletronic.com
 */
public class Converter {

    /**
     * usage: java -jar syxconverter.jar mr9090.hex mr9090.syx
     */
    public static void main(String[] args) {
        if(args.length < 2) {
            System.out.println("usage: java -jar syxconverter.jar mr9090.hex mr9090.syx");
            return;
        }

        BufferedReader in = null;
        BufferedOutputStream out = null;
        try {
            in = new BufferedReader(new FileReader(args[0]));
            out = new BufferedOutputStream(new FileOutputStream(args[1]));
            new Converter().convertFile(in, out);
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            if(in != null)
                try { in.close(); } catch (Exception ee) {}
            if(out != null)
                try { out.flush(); out.close(); } catch (Exception ee) {}
        }
    }

    // address extension
    private int addre = 0;

    public void convertFile(BufferedReader in, BufferedOutputStream out) throws Exception {
        SysexChunk chunk = null;
        INHX32Row row;

        while((row = readRow(in)) != null) {
            System.out.println(row);
            switch(row.tt) {
                case 0:
                    // data record
                    if(chunk == null) {
                        // first or continued chunk
                        chunk = new SysexChunk();
                        chunk.addr = row.addr + addre;
                    }

                    // chunk needs data,
                    //  check if chunk & file row addresses match
                    if((chunk.addr + chunk.count) != (row.addr + addre)) {
                        // non-continuous block, pad with zeroes to make a full 64 byte chunk
                        System.out.println(
                            "non-continuous block at 0x" + Integer.toHexString(row.addr + addre) + ", padding between 0x" +
                            Integer.toHexString(chunk.addr + chunk.count) + " - 0x" +
                            Integer.toHexString(row.addr + addre)
                        );
                        chunk.pad(row.addr + addre);
                    }

                    // fill chunk with row data
                    for(int i=0; i<row.bb; i++) {
                        if(chunk.count == 64) {
                            // chunk finished, write to file and create new
                            chunk.write(out);
                            chunk.count = 0;
                            chunk.addr += 64 + addre;
                        }
                        chunk.bytes[chunk.count] = row.bytes[i];
                        chunk.count++;
                    }

                    break;
                case 4:
                    // extended address record, get extension
                    addre = (row.bytes[1] << 16) + (row.bytes[0] << 8);
                    break;
                case 1:
                    // end of file record
                    // just ignore
                    break;
            }
        }
        // end of file, do we have unwritten data ?
        if(chunk != null && chunk.count > 0) {
            chunk.pad(Integer.MAX_VALUE);       // pad a full 64 byte chunk
            chunk.write(out);
        }
        System.out.println("Sysex file written successfully.");
    }

    /**
     * Reads row from input file and parses the data to INHX32Row object.
     *
     * :BBAAAATTHHHH...HHHHCC
     */
    private INHX32Row readRow(BufferedReader in) throws Exception {
        String line = in.readLine();
        if(line == null)
            return null;
        INHX32Row row = new INHX32Row();
        row.bb = Integer.parseInt(line.substring(1, 3), 16);
        row.addr = Integer.parseInt(line.substring(3, 5), 16) << 8; // ah
        row.addr += Integer.parseInt(line.substring(5, 7), 16);     // al
        row.tt = Integer.parseInt(line.substring(7, 9), 16);
        row.bytes = new int[row.bb];

        for(int i=0; i<row.bb; i++) {
            row.bytes[i] = Integer.parseInt(line.substring(9 + (i * 2), 11 + (i * 2)), 16);
        }
        return row;
    }

    class INHX32Row {
        int bb;
        int addr;
        int tt;
        int[] bytes;

        public String toString() {
            if(tt == 4)
                return "input: address extension 0x" + Integer.toHexString((bytes[0] << 8) + bytes[1]);
            else if(tt == 0)
                return "input: " + bb + " bytes at address 0x" + Integer.toHexString(addr + addre);
            else if(tt == 1)
                return "input: file record, eof?";
            else
                return "input: unknown record type " + tt;
        }
    }

    class SysexChunk {
        int addr;
        int count;
        int[] bytes = new int[64];

        // pads the remaining bytes with zeroes until raddr to make a full chunk
        public void pad(int raddr) {
            while(count < 64 && (addr + count) < raddr) {
                bytes[count++] = 0;
            }
        }

        /**
         * Writes sysex chunk to output file.
         *
         * F7 71 55 17 7F AH AL &lt;64 bytes of midi encoded (BYTE%gt;&gt;7 BYTE&0x7F) data&gt; F7<br/>
         * <br/>
         *  AH = program memory address &gt;&gt; 7 & 0x7f<br/>
         *  AL = program memory address & 0x7f<br/>
         */
        public void write(BufferedOutputStream out) throws Exception {
            if(addr < 0x4000) {
                // addresses in program memory
                if((addr & 0x03) != 0) {
                    throw new Exception("oh no, address not in 8 byte boundary: " + Integer.toHexString(addr));
                }

                final byte[] header = { (byte)0xf0, 0x71, 0x55, 0x17, 0x7f };
                out.write(header);

                // 00000000 10000000

                // 00000000 00000001 AH

                out.write(((addr >> 7) & 0x7f));  // AH >> 7 & 0x7f
                out.write((addr & 0x7f));         // AH & 0x7f
                for(int i=0; i<count; i++) {
                    out.write((bytes[i] >> 7) & 0x7f);
                    out.write((bytes[i] & 0x7f));
                }
                System.out.println(" 64 bytes at 0x" + Integer.toHexString(addr) + " written.");
                out.write(0xf7);    // sysex end
            } else {
                System.out.println(" ignoring data at 0x" + Integer.toHexString(addr) + ", probably eeprom or config data?");
            }
        }

    }

} // class
